#include <stdio.h>           /* for stderr, fprintf() */
#include <stdlib.h>
#include <errno.h>
#include <syslog.h>          /* for LOG_ALERT, syslog() */
#include <sys/types.h>       /* for uid_t */
#include <pwd.h>             /* for struct passwd, getpwnam */
#include <unistd.h>          /* for seteuid() */
#include <sys/time.h>      /* for stuff in resource.h */
#include <sys/resource.h>  /* for struct rlimit, RLIM_INFINITY, RLIMIT_NPROC,
			      setrlimit() */
#include <signal.h>
#include <comdef.h>
#include "inetcis_common.h"  /* for inetcis_status,
			      INETCIS_STATUS_FAILURE,
			      INETCIS_STATUS_SUCCESS */

inetcis_status_type become_user (char *user_name, char *prog_id,
				 char *output_line_terminator)
    {
    struct passwd *user_passwd_entry_ptr;
    uid_t new_uid;
    int status;

    /*
     * Get uid for user user_name.  If this doesn't work, we cannot
     * continue because we would end up opening a xpad as root, which
     * screw up whoever was already using that xpad.
     */

    user_passwd_entry_ptr = getpwnam (user_name);
    if (user_passwd_entry_ptr == GETPWNAM_ERROR)
	{
	syslog (LOG_ALERT, "%s %d - alert - function getpwnam caused system"
		" error `%m' trying to get passwd file entry for user %s",
		__FILE__, __LINE__, user_name);
	fprintf (stderr, "? %sUSR Unable to find user %s%s", prog_id,
		 user_name, output_line_terminator);
	return INETCIS_STATUS_FAILURE;
	}
    new_uid = user_passwd_entry_ptr->pw_uid;

    /*
     * Set the effective uid for this process to be the uid of the
     * user user_name.  If this doesn't work we cannot continue for the
     * same reason as specified for getpwnam.
     */

    status = seteuid (new_uid);
    if (status == SETEUID_ERROR)
	{
	syslog (LOG_ALERT, "%s %d - alert - function seteuid caused system"
		" error `%m' trying to set the effective user ID of this"
		" process to %d", __FILE__, __LINE__, (int) new_uid);
	fprintf (stderr, "? %sUID Unable to become user %s%s", prog_id,
		 user_name, output_line_terminator);
	return INETCIS_STATUS_FAILURE;
	}

    syslog (LOG_DEBUG, "%s %d - debug - became user %s", __FILE__, __LINE__,
	    user_name);

    /*
     * Change working directory to home directory of the user as who we are
     * now running.  This is mainly done so that core files can successfully
     * be written to a directory that we have write access to.  We don't
     * really need to exit or return failure if it doesn't work.  Just logging
     * the error will be enough.
     */
    
    status = chdir (user_passwd_entry_ptr->pw_dir);
    if (status == CHDIR_ERROR)
	{
	syslog (LOG_ERR, "%s %d - err - function chdir caused system error"
		" `%m' trying to change to home directory %s of user %s",
		__FILE__, __LINE__, user_passwd_entry_ptr->pw_dir, user_name);
	}

    return INETCIS_STATUS_SUCCESS;
    }
    /* end of become_user */







void increase_process_limit ()
    {
    struct rlimit process_limit_struct;
    int status;

    /*
     * Set limits for number of processes to RLIM_INFINITY.  If this
     * doesn't work we can still continue in hopes that the soft limit
     * hasn't been reached yet.
     */

    process_limit_struct.rlim_cur = RLIM_INFINITY;
    process_limit_struct.rlim_max = RLIM_INFINITY;
    status = setrlimit (RLIMIT_NPROC, &process_limit_struct);
    if (status == SETRLIMIT_ERROR)
	{
	syslog (LOG_ALERT, "%s %d - alert - function setrlimit caused"
		" system error `%m' trying to set process limit to %d"
		" (RLIM_INFINITY)", __FILE__, __LINE__, RLIM_INFINITY);
	}
    else
	{
	syslog (LOG_DEBUG, "%s %d - debug - set process limit to %d"
		" which is RLIM_INFINITY", __FILE__, __LINE__, RLIM_INFINITY);
	}
    }



int Read_from_X25_Device (int fd, char *buffer, int number_of_bytes_to_read)
    {
    const int MAX_BUFFER_SIZE = 4096;
    int number_of_bytes_read;
    char temp_buffer[MAX_BUFFER_SIZE + 1];

    number_of_bytes_read = read (fd, buffer, number_of_bytes_to_read);

    if (READ_ERROR == number_of_bytes_read)
	{
	if (errno == EWOULDBLOCK)
	    {
	    syslog (LOG_DEBUG, "%s %d - debug - function read caused system"
		    " error `%m' trying to read from X.25 device",
		    __FILE__, __LINE__);
	    number_of_bytes_read = 0;
	    }
	else
	    {
	    syslog (LOG_ERR, "%s %d - err - function read caused system"
		    " error `%m' trying to read from X.25 device",
		    __FILE__, __LINE__);
	    exit (EXIT_FAILURE);
	    }
	} /* if */
    else if (number_of_bytes_read == READ_EOF)
	{
	syslog (LOG_INFO, "%s %d - info - read EOF from X.25 device",
		__FILE__, __LINE__);
	exit (EXIT_SUCCESS);
	}
    else
	{
	memcpy (temp_buffer, buffer, number_of_bytes_read);
	temp_buffer[number_of_bytes_read] = EOS;

#ifdef LOG_EVERYTHING
	{
	char codes[MAX_BUFFER_SIZE * 4 + 1];
	Convert_Characters_in_String_to_ASCII_Codes (temp_buffer, codes);
	syslog (LOG_DEBUG, "Read %d bytes from device: %s Codes: %s",
		number_of_bytes_read, temp_buffer, codes);
	}
#endif

	}  /* if (READ_ERROR == number_of_bytes_read) */
	
    return number_of_bytes_read;

    } /* Read_from_X25_Device */





void Strip_Leading_CRs_And_LFs (char *s_in, char *s_out, int size_in,
                                int *size_out)

/*
 * Functional Description:
 *
 *     Removes any carriage returns or line feeds found at the beginning of
 *     s_in.  It processes s_in until it finds a character other than a
 *     carriage return or line feed.
 *
 * Formal Parameters:
 *
 *     s_in (read only)incoming string which is not NULL terminated
 *     s_out    - outgoing string which is not NULL terminated
 *     size_in  - number of characters in s_in
 *     size_out - number of characters in the resulting string s_out
 *
 * Implicit Parameters:
 *
 *     none
 *
 * Return Values:
 *
 *     s_out, size_out
 *
 * Side Effects:
 *
 *     none
 *
 */

    {
    int i = 0;
    int j = 0;

    while (s_in[i] == '\r' || s_in[i] == '\n')
            i++;

    while (i < size_in)
        {
        s_out[j++] = s_in[i++];
        }

    *size_out = j;
    }




void read_user_id_prompt (int file_descriptor)
    {
    const int max_buffer_size = 4096;
    char stripped_buffer[max_buffer_size];
    int character_count;
    char buffer[max_buffer_size];
    int stripped_buffer_length = 0;
    fd_set select_readers;
    int system_status;
    struct timeval timeout_period = {30L /* seconds */, 0L /* microseconds */};

    do
        {
	FD_ZERO (&select_readers);
	FD_SET (file_descriptor, &select_readers);
	system_status = select (file_descriptor + 1, &select_readers,
				(fd_set *) NULL,
				(fd_set *) NULL, &timeout_period);
	if (system_status == SELECT_ERROR)
	    {
	    syslog (LOG_ERR, "%s %d - err - function select caused system"
		    " error %m trying to wait for user id prompt on X.25"
		    " device", __FILE__, __LINE__);
	    exit (EXIT_FAILURE);
	    }
	else if (system_status == SELECT_TIMEOUT)
	    {
	    syslog (LOG_ERR, "%s %d - err - function select timed out waiting"
		    " for user id prompt on X.25 device", __FILE__, __LINE__);
	    exit (EXIT_FAILURE);
	    }
	else if (FD_ISSET (file_descriptor, &select_readers))
	    {
	    character_count = Read_from_X25_Device (file_descriptor, buffer,
						    max_buffer_size);
	    Strip_Leading_CRs_And_LFs (buffer, stripped_buffer,
				       character_count,
				       &stripped_buffer_length);
	    }
	else
	    {
	    syslog (LOG_ERR, "%s %d - err - unknown problem - select returned"
		    " normally, but did not set file descriptor in descriptor"
		    " set", __FILE__, __LINE__);
	    }
        } 
    while (stripped_buffer_length <= 0 ||
	   0 != strncmp ("User ID: ", stripped_buffer, 
			 stripped_buffer_length));

    }  /* read_user_id_prompt() */



void segmentation_fault_handler (int signal_number)
    {
    int system_status;
    syslog (LOG_ALERT, "%s %d - alert - caught segmentation fault signal,"
	    " SEGV - signal number = %d", __FILE__, __LINE__, signal_number);
    syslog (LOG_NOTICE, "%s %d - notice - raising abort signal ABRT to"
	    " create a core dump", __FILE__, __LINE__);
    system_status = raise (SIGABRT);
    if (system_status == RAISE_ERROR)
	{
	syslog (LOG_ERR, "%s %d - err - function raise caused system error"
		" `%m' trying to raise abort signal SIGABRT", __FILE__,
		__LINE__);
	}

    /*
     * If execution gets to this point, probably several things have failed.
     */

    exit (EXIT_FAILURE);
    }

