/*
 * Copyright (c) 1995 by CompuServe Incorporated, Columbus, Ohio
 *
 * Facility:  libinetcis.a
 *
 * Abstract:
 *
 *     This module contains the functions for performing input/output on an
 *     X.25 device (tpad).  It handles errors and can cause a process to exit.
 *     So, the program that uses this module should have an exit handler
 *     defined with atexit().  It was designed for inetcis processes.
 *
 * Environment:  BSDI's BSD/OS 1.1 (POSIX), Requires xpad/tpad interfaces to
 *               be available
 *
 * Author:  William L. Chatfield (BCHATFIELD), 12-Jul-95
 *
 * Revision History:
 *
 * 1.0  William L. Chatfield (BCHATFIELD), 12-Jul-95
 *      - Initial revision
 */

#ifndef X25IO_H
#define X25IO_H

/*
 * External Interfaces
 */
#include <comdef.h>                  /* CompuServe definitions */

/*
 * Constants, structures, and typedefs
 */
#define NUM_TPADS 128
#define GET_TPAD_ERROR_NO_AVAILABLE_DEVICES -1

/*
 * External data
 */

/*
 * Function prototypes
 */


extern int Open_X25_Device (char *device_name);
/*
 * Functional Description:
 *
 *      This function finds an available X.25 device (tpad) on the machine and
 *      opens it.
 *
 * Formal Parameters:
 *
 *      tpad_name (Write only)
 *          - This is the variable in which the name of the device that was
 *            opened is returned.
 *
 * Implicit Parameters:  None
 *
 * Return Value:
 *
 *      If a device could be found, its file descriptor is returned.
 *      If no available device was found GET_TPAD_ERROR_NO_AVAILABLE_DEVICE
 *      is returned.
 *
 * Side Effects:  None
 *
 */

extern int Read_From_X25_Device (int fd, char *buffer,
				 int number_of_bytes_to_read,
				 int *number_of_bytes_read);
/*
 * Functional Description:
 *
 *      This function performs a read() on the file descriptor that is passed
 *      to it.  It checks for errors and handles them correctly.
 *
 * Requirements:
 *
 *      1. There must be a syslog type log file open so that syslog calls will
 *         work.
 *      2. fd must be a valid open file descriptor for an X.25 device.
 *      3. fd should be in non-blocking mode.
 *
 * Formal Parameters:
 *
 *      fd (Read only)
 *          - The file descriptor of the X.25 device from which data will be
 *            read.
 *      buffer (Write only)
 *          - A pointer to the area of memory into which the data read will be
 *            put.  The size of this area of memory should be equal to
 *            number_of_bytes_to_read or greater.
 *      number_of_bytes_to_read (Read only)
 *          - The number of bytes to read from the X.25 device.  This number
 *            should be equal to or less than the size of the memory pointed to
 *            by buffer.
 *
 * Implicit Parameters:  None
 *
 * Return Values:
 *
 *      The number of bytes read is returned.
 *
 * Side Effects:  None
 *
 */

extern int Write_To_X25_Device (int fd, char *buffer,
				int number_of_bytes_to_write,
				int *number_of_bytes_written);
/*
 * Functional Description:
 *
 *      This function performs a write() on the file descriptor that is passed
 *      to it.  It checks for errors and handles them correctly.
 *
 * Requirements:
 *
 *      1. There must be a syslog type log file open so that syslog calls will
 *         work.
 *      2. fd must be a valid open file descriptor for an X.25 device.
 *      3. fd should be in non-blocking mode.
 *
 * Formal Parameters:
 *
 *      fd (Read only)
 *          - The file descriptor of the X.25 device from which data will be
 *            read.
 *      buffer (Read only)
 *          - A pointer to the area of memory from which the data will be
 *            taken.  The size of this area of memory should be equal to
 *            number_of_bytes_to_read or greater.
 *      number_of_bytes_to_write (Read only)
 *          - The number of bytes to write to the X.25 device.  This number
 *            should be equal to or less than the size of the memory pointed to
 *            by buffer.
 *
 * Implicit Parameters:  None
 *
 * Return Values:
 *
 *      The number of bytes written is returned.
 *
 * Side Effects:  None
 *
 */

extern int Write_All_To_X25_Device (int file_descriptor, char *buffer_ptr,
				    int number_of_bytes_to_write);
/*
 * Functional Description:
 *
 *      This function writes all the bytes specified and does not return until
 *      it has written all of them or an error occurs.
 *
 * Formal Parameters:
 *
 *      file_descriptor (Read only)
 *          - descriptor to write to
 *      buffer_ptr (Read only)
 *          - pointer to the buffer from which the bytes written will be taken
 *      number_of_bytes_to_write (Read only)
 *          - number of bytes to write
 *
 * Implicit Parameters:  None
 *
 * Return Value:
 *
 *      None
 *
 * Side Effects:  None
 *
 */

extern int Write_All_To_X25_Device_EonF (int fd, char *buffer_ptr,
					 int number_to_write);
extern void Log_X25_Device_Stats ();
extern void Start_Producing_X25_Flow_Stats ();
extern int Reopen_X25_Device (int *x25_fd, char *x25_device_name);
extern int Read_PAD_Prompt (int x25_fd);
extern int Test_X25_Device (int x25_fd);
extern int Refresh_X25_Device (int *x25_fd, char *x25_device_name);
extern int Rotor_Open_X25_Device (int *x25_fd, char *x25_device_name);
extern int Ask_PAD_For_Prompt (int x25_fd);
extern void Log_All_Links_In_Use_Error (char *file, int line, char *operation);
extern void Read_User_ID_Prompt (int fd);

#endif
