/*
 * inetcis.c
 */

/*
 * System includes
 */
#include <sys/types.h>		/* select */
#include <sys/wait.h>		/* wait */
#include <sys/socket.h>		/* socket */
#include <netinet/in.h>		/* struct sockaddr_in */
#include <netdb.h>		/* gethostbyname, herror, and struct hostent */
#include <arpa/inet.h>
#include <unistd.h>		/* STDIN_FILENO, STDOUT_FILENO */
#include <sys/time.h>		/* struct timeval */
#include <errno.h>		/* errno */
#include <sys/fcntl.h>		/* F_SETFL */

/*
 * ANSI C includes
 */
#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>
#include <string.h>             /* bzero */

/*
 * Local includes
 */
#include <ourhdr.h>		/* tty_raw, tty_reset */
#include <poserr.h>

/*
 * Holy Macro!
 */
#define	min(a,b) ((a) < (b) ? (a) : (b))
#define	max(a,b) ((a) > (b) ? (a) : (b))
#define BUFFER_SIZE 1024
#define TTY_RAW_ERROR -1
#define TTY_RESET_ERROR -1
#define SERVICE_NAME "inetcis"
#define ESCAPE_CHARACTER 0x01f
#define ESCAPE_CHARACTER_STRING "^_"

/*
 * Typedefs
 */
typedef struct in_addr **InAddrPointerArrayType;
typedef enum {
  FAILURE, SUCCESS
}               StatusType;

/*
 * Global variables
 */
const char      Escape_Character = '';
char            Program_Name[256];
int             Socket_To_CompuServe;
u_short         Default_Port_Number = 4144;	/* must use htons() on this
						 * value */

/*
 * Function prototypes.
 */
void            Build_Server_Socket_Address (struct sockaddr_in *, char *);
void            Process_Inputs ();
void            Process_Incoming_Data_From_CompuServe ();
void            Process_Incoming_Data_From_stdin ();
void            Exit_inetcis (const int, int, const char *);
void            Send_Data_Through_Socket (int, char *, int);
int             Remove_Escape_And_Following_Characters (char *, int *);
void            Receive_File (void);
void            Exit_Handler (void);
void 
Get_Server_Address_Pointer_Array (char *server_name,
				  InAddrPointerArrayType
				  * potential_server_list);
StatusType 
Try_Connects (int socket, struct sockaddr_in * server_socket_address,
	      InAddrPointerArrayType potential_server_list);

void 
main (int argc, char **argv)
{
  struct sockaddr_in server_socket_address;
  char            server_machine_name_string[256];
  InAddrPointerArrayType potential_servers;
  StatusType      status = SUCCESS;

  atexit (Exit_Handler);
  /*
   * Turn off buffering with stdio.
   */
  (void) setbuf (stdin, (char *) NULL);
  (void) setbuf (stdout, (char *) NULL);

  /*
   * Process argument list.
   */
  strcpy (Program_Name, argv[0]);
  if (argc == 2) {
    strcpy (server_machine_name_string, argv[1]);
  } else {
    strcpy (server_machine_name_string, "gateway.compuserve.com");
  }

  printf ("Escape character is %s\n", ESCAPE_CHARACTER_STRING);

  /*
   * Define a socket to be used for the CompuServe connection.
   */
  Socket_To_CompuServe = socket (AF_INET, SOCK_STREAM, IPPROTO_IP);
  if (Socket_To_CompuServe == SOCKET_ERROR) {
    Exit_inetcis (EXIT_FAILURE, errno, "socket call");
  }
  Build_Server_Socket_Address (&server_socket_address, SERVICE_NAME);

  if (isalpha (server_machine_name_string[0])) {
    Get_Server_Address_Pointer_Array (server_machine_name_string,
				      &potential_servers);
    status = Try_Connects (Socket_To_CompuServe,
			   &server_socket_address,
			   potential_servers);
  } else {
    unsigned long single_address;

    single_address = inet_addr (server_machine_name_string);
    if (single_address == INADDR_NONE) {
      puts ("Malformed numeric address\r");
      status = FAILURE;
    } else {
      server_socket_address.sin_addr.s_addr = single_address;
      printf ("Attempting to connect to %s", server_machine_name_string);
      if (connect (Socket_To_CompuServe,
		   (struct sockaddr *) &server_socket_address,
		   sizeof (struct sockaddr_in)) == CONNECT_ERROR) {
	printf (" - failed: %s\n\n", strerror(errno));
	status = FAILURE;
      } else {
	printf (" - success\n\n");
      }
    }
  }

  if (status == FAILURE) {
    puts ("Could not connect to any server.\r");
    exit (EXIT_FAILURE);
  }
  Process_Inputs ();

  close (Socket_To_CompuServe);

}				/* main */


void 
Build_Server_Socket_Address (struct sockaddr_in * socket_address,
			     char *service_name)
{
  struct servent *server_service_info_ptr;

  /*
   * Initialize socket_address which is a `struct sockaddr_in'.
   */
  bzero (socket_address, sizeof (struct sockaddr_in));

  /*
   * Put the socket family name in the socket_address structure.
   */
  socket_address->sin_family = AF_INET;

  /*
   * Put port number in socket_address.
   */
  server_service_info_ptr = getservbyname (service_name, (char *) NULL);
  if (server_service_info_ptr == GETSERVBYNAME_ERROR) {
#if 0
    /* This is annoying. */
    fprintf (stderr, "Unable to get port number for service %s, trying"
	     " %d.\n", service_name, (int) Default_Port_Number);
#endif
    socket_address->sin_port = htons (Default_Port_Number);
  } else {
    socket_address->sin_port = server_service_info_ptr->s_port;
  }

}				/* Assign_Socket_Address_Info */



void 
Process_Inputs ()
{
  int             number_of_ready_descriptors;
  fd_set          readers;
  int             max_fd;
  struct timeval  select_timeout_value;

  /*
   * Initialize variables
   */
  max_fd = max (STDIN_FILENO, Socket_To_CompuServe);
  select_timeout_value.tv_sec = 120;
  select_timeout_value.tv_usec = 0;

  /* This makes STDOUT_FILENO non-blocking also! */
  set_fl (STDIN_FILENO, O_NONBLOCK);
  set_fl (Socket_To_CompuServe, O_NONBLOCK);

  /*
   * Put the tty for stdin in raw mode so that the CompuServe side will
   * handle terminal control.
   */
  if (isatty (STDIN_FILENO)) {
    if (TTY_RAW_ERROR == tty_raw (STDIN_FILENO)) {
      fprintf (stderr, "%s: tty_raw caused error: %s\r\n", Program_Name,
	       strerror (errno));
      exit (EXIT_FAILURE);
    }
  }
  while (1) {
    /*
     * Setup the file descriptor set used in the select system call.
     */
    FD_ZERO (&readers);
    FD_ZERO (&readers);
    FD_SET (Socket_To_CompuServe, &readers);
    FD_SET (STDIN_FILENO, &readers);

    /*
     * Wait until something is there to read.
     */
    number_of_ready_descriptors = select (max_fd + 1, &readers,
					  (fd_set *) NULL,
					  (fd_set *) NULL,
					  (struct timeval *) NULL);
    switch (number_of_ready_descriptors) {
    case SELECT_ERROR:
      fprintf (stderr, "%s: select caused error `%s'\n", Program_Name,
	       strerror (errno));
      fprintf (stderr, "%s: time out secs: %d, usec: %d", Program_Name,
	       select_timeout_value.tv_sec,
	       select_timeout_value.tv_usec);
      exit (EXIT_FAILURE);
      break;
    case SELECT_TIMEOUT:
      fprintf (stderr, "%s: select timed out waiting for data\n",
	       Program_Name);
      fprintf (stderr, "%d %d %d %d %d %d\n",
	       readers.fds_bits[0] && 0x01,
	       readers.fds_bits[0] && 0x02,
	       readers.fds_bits[0] && 0x04,
	       readers.fds_bits[0] && 0x08,
	       readers.fds_bits[0] && 0x10,
	       readers.fds_bits[0] && 0x12);
      break;
    case 2:
      Process_Incoming_Data_From_CompuServe (Socket_To_CompuServe);
      Process_Incoming_Data_From_stdin (Socket_To_CompuServe);
      break;
    case 1:
      if (FD_ISSET (Socket_To_CompuServe, &readers)) {
	Process_Incoming_Data_From_CompuServe (Socket_To_CompuServe);
      } else if (FD_ISSET (STDIN_FILENO, &readers)) {
	Process_Incoming_Data_From_stdin (Socket_To_CompuServe);
      }
      break;
    default:
      fprintf (stderr, "%s: something bad happened\n", Program_Name);
    }				/* switch */

  }				/* while */

  /* This point should be unreachable. */

}				/* Process_Input */


void 
Process_Incoming_Data_From_CompuServe ()
{
  int             number_of_bytes_read;
  char            buffer[BUFFER_SIZE];

  number_of_bytes_read = read (Socket_To_CompuServe, buffer, BUFFER_SIZE);
  if (READ_ERROR == number_of_bytes_read) {
    if (errno != EAGAIN) {
      Exit_inetcis (EXIT_FAILURE, errno, "read from socket");
    }
  } else if (READ_EOF == number_of_bytes_read) {
    write (STDOUT_FILENO, "\r\n", 2);
    Exit_inetcis (EXIT_SUCCESS, 0, 0);
  } else {			/* There is data.  Give it to the user. */
    int             number_written = write (STDOUT_FILENO, buffer,
					    (size_t) number_of_bytes_read);
    if (WRITE_ERROR == number_written) {
      if (errno != EAGAIN) {
	Exit_inetcis (EXIT_FAILURE, errno, "write to stdout");
      }
    }
  }
}				/* Process_Incoming_Data_From_CompuServe */


void 
Process_Incoming_Data_From_stdin ()
{
  char            buffer[BUFFER_SIZE];
  int             number_of_bytes_read;

  /*
   * Read from stdin and write to socket.
   */
  number_of_bytes_read = read (STDIN_FILENO, buffer, BUFFER_SIZE);
  if (READ_ERROR == number_of_bytes_read) {
    if (errno != EAGAIN) {
      Exit_inetcis (EXIT_FAILURE, errno, "read from stdin");
    }
  } else {

    /* Respond to telnet escape sequence.  This might someday trigger
     a file download or a command prompt interface. */

    if (buffer[0] == ESCAPE_CHARACTER)
      exit (EXIT_SUCCESS);

    Send_Data_Through_Socket (Socket_To_CompuServe, buffer,
			      number_of_bytes_read);
  }
}				/* Process_Incoming_Data_From_stdin */


void 
Exit_inetcis (const int exit_type, int error_number,
	      const char *program_position)
{
  char            message[256];

  if (exit_type == EXIT_FAILURE) {
    strcpy (message, Program_Name);
    strcat (message, " at ");
    strcat (message, program_position);
    fprintf (stderr, "\r\n%s: %s\r\n", message, strerror (error_number));
  }
  if (exit_type == EXIT_FAILURE) {
    exit (EXIT_FAILURE);
  } else {
    exit (EXIT_SUCCESS);
  }
}

void 
Send_Data_Through_Socket (int sproket, char *buffer,
			  int total_number_of_bytes_to_send)
{
  int             number_of_bytes_left_to_send = total_number_of_bytes_to_send;
  int             number_of_bytes_sent = 0;

  while (number_of_bytes_left_to_send != 0) {
    number_of_bytes_sent = send (sproket, buffer,
				 number_of_bytes_left_to_send,
				 EMPTY_SOCKET_FLAGS);
    if (SEND_ERROR == number_of_bytes_sent) {
      if (errno == EAGAIN) {
	number_of_bytes_sent = 0;
      } else {
	Exit_inetcis (EXIT_FAILURE, errno, "socket sent");
      }
    }
    number_of_bytes_left_to_send -= number_of_bytes_sent;
  }				/* while */
}				/* Send_Data_Through_Socket */



void 
Exit_Handler (void)
{
  if (isatty (STDIN_FILENO)) {
    if (TTY_RESET_ERROR == tty_reset (STDIN_FILENO)) {
      perror (Program_Name);
    }
  }
  clr_fl (STDIN_FILENO, O_NONBLOCK);
  clr_fl (Socket_To_CompuServe, O_NONBLOCK);

  close (Socket_To_CompuServe);
}

void 
Get_Server_Address_Pointer_Array (char *server_machine_name_string,
				  InAddrPointerArrayType
				  * address_ptr_array)
{
  struct hostent *server_hostent_ptr;

  server_hostent_ptr = gethostbyname (server_machine_name_string);
  /*
   * Check for errors in getting server_hostent_ptr.
   */
  if (server_hostent_ptr == NULL) {
    fprintf (stderr, "Unable to get host name for host %s.\n",
	     server_machine_name_string);
    herror (Program_Name);
    exit (EXIT_FAILURE);
  }
  
  /*
   * Point *address_ptr_array at the potential server list in the `struct
   * hostent' structure pointed to by server_hostent_ptr.
   */
  *address_ptr_array = (InAddrPointerArrayType)
    server_hostent_ptr->h_addr_list;
}				/* Get_Server_Address_Pointer_Array */

StatusType 
Try_Connects (int socket,
	      struct sockaddr_in * server_socket_address_ptr,
	      InAddrPointerArrayType potential_server_list)
{
  int             system_status;
  int             i = 0;
  char           *dotted_decimal_address;

  while (potential_server_list[i] != NULL) {

    /*
     * Copy the server information for this list entry in to the socket
     * address structure which will be used for the connect() call.
     */
    server_socket_address_ptr->sin_addr.s_addr =
      potential_server_list[i]->s_addr;
    
    dotted_decimal_address = inet_ntoa (server_socket_address_ptr->sin_addr);
    printf ("Attempting to connect to %s", dotted_decimal_address);

    system_status = connect (socket,
			     (struct sockaddr *) server_socket_address_ptr,
			     sizeof (struct sockaddr_in));
    
    if (system_status != CONNECT_ERROR) {
      printf (" - success\n\n");
      return SUCCESS;
    } else {
      printf (" - failed: %s\n\n", strerror (errno));
    }
    
    i++;
  }


  return FAILURE;
}
