/*
 * Copyright (c) 1995 by CompuServe Incorporated, Columbus, Ohio
 * 
 * Facility:  libinetcis.a
 * 
 * Abstract:
 * 
 * This module contains the I/O operations that can be performed on a socket in
 * inetcis programs.  The error reporting, Socket_Is_Alive(), only works when
 * the functions in this module are used with one socket.
 * 
 * Environment:  BSDI's BSD/OS 1.1 (POSIX)
 * 
 * Author:  William L. Chatfield (BCHATFIELD), 14-Jul-95
 * 
 * Revision History:
 * 
 * 1.0  William L. Chatfield (BCHATFIELD), 14-Jul-95 - Initial revision
 */

/*
 * POSIX Interfaces
 */
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <syslog.h>
#include <errno.h>
#include <netdb.h>

/*
 * ANSI C Interfaces
 */
#include <stdlib.h>

#include <comdef.h>		/* CompuServe definitions */

#include <log.h>
#include <poserr.h>
#include <inetcis-types.h>

/*
 * Variables in system modules that should have been defined in system
 * system header files.
 */
extern int h_errno;


/*
 * Module-wide definitions */
#define MODULE_BUFFER_SIZE 4096
static bool     socket_is_alive = TRUE;	/* socket is passed open on
					 * descriptor 0 */
static double   old_number_of_buffers_read = 0.0;
static double   number_of_buffers_read = 0.0;
static double   average_read_buffer_size = 0.0;
static double   number_of_read_eagain_errors = 0.0;
static double   old_number_of_buffers_written = 0.0;
static double   number_of_buffers_written = 0.0;
static double   average_write_buffer_size = 0.0;
static double   number_of_write_eagain_errors = 0.0;
static int      produce_flow_stats = 0;
static char     module_buffer[MODULE_BUFFER_SIZE];

/*
 * Global data
 */

/*
 * Table of contents
 */

int
Write_To_Socket (int fd, char *buffer_ptr, int number_of_bytes_to_write,
		 int *number_of_bytes_written)
/*
 * Functional Description:
 * 
 * This function writes to a socket and handles any errors.
 * 
 * Formal Parameters:
 * 
 * socket_file_descriptor (Read only) - The descriptor of the socket to which to
 * write. buffer_ptr (Write only) - Pointer to buffer to which data will be
 * taken. number_of_bytes_to_write (Read only) - Number of bytes to write.
 * 
 * Implicit Parameters:  None
 * 
 * Return Value:
 * 
 * The number of bytes written is returned.
 * 
 * Side Effects:  The module-wide variable socket_is_alive is set.
 * 
 */
{
  int             n;

  if (number_of_bytes_written == NULL || buffer_ptr == NULL)
    return FAILURE;

  n = write (fd, buffer_ptr, (size_t) number_of_bytes_to_write);

  if (n == WRITE_ERROR) { 
    n = 0;
    if (errno == EAGAIN) {
      Log_Hidden_System_Error (LOG_LEVEL_DEBUG0, __FILE__, __LINE__, errno,
			       "socket", "write", "Write would block");
      number_of_write_eagain_errors++;
    } else if (errno == EHOSTUNREACH || errno == ECONNRESET) {
      
      /* EHOSTUNREACH is the `No route to host' error that happens
	 probably when a PPP user shuts off his computer without
	 disconnecting cleanly.

	 ECONNRESET is the `Connection reset by peer' error that seems
	 to be a `normal' connection termination message.

	 Both messages are part of the normal operation of the system
	 and should only be logged as debug messages.  They happen
	 quite often.  If they are logged as errors they put lots of
	 meaningless messages in the syslog log file on production
	 machines. */

      Log_Hidden_System_Error (LOG_LEVEL_DEBUG0, __FILE__, __LINE__, errno,
			       "socket", "write",
			       "Connection was terminated incorrectly by "
			       "client");
      return FAILURE;
      
    } else {
      Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
				 "socket", "write", "WRT",
				 "Cannot write to client");
      return FAILURE;
    }
  }

  /* Calculate statistics */
  if (produce_flow_stats) {
    old_number_of_buffers_written = number_of_buffers_written;
    number_of_buffers_written++;
    average_write_buffer_size = (average_write_buffer_size *
				 old_number_of_buffers_written + n) /
				   number_of_buffers_written;
  }

  *number_of_bytes_written = n;
  return SUCCESS;
}


int 
Read_From_Socket (int fd, char *buffer_ptr, int number_of_bytes_to_read,
		  int *number_of_bytes_read)
/*
 * Functional Description:
 * 
 * This function reads from a socket and handles any errors and EOF contitions.
 * 
 * Formal Parameters:
 * 
 * fd (Read only) - The socket's file descriptor.
 * buffer_ptr (Read only) - Pointer to the buffer in which data is put.
 * number_of_bytes_to_read (Read only) - Number of bytes to read.
 * 
 * Implicit Parameters:  None
 * 
 * Return Value:
 * 
 * The number of bytes read is returned.
 * 
 * Side Effects:  The module-wide variable socket_is_alive is set.
 */
{
  int             num;

  if (number_of_bytes_read == NULL || buffer_ptr == NULL)
    return FAILURE;

  num = read (fd, buffer_ptr, number_of_bytes_to_read);
  if (num == READ_ERROR) {
    num = 0;
    if (errno == EAGAIN) {
      Log_Hidden_System_Error (LOG_LEVEL_DEBUG0, __FILE__, __LINE__, errno,
			       "socket", "read", "Nothing to read");
      number_of_read_eagain_errors++;

    } else if (errno == EHOSTUNREACH || errno == ECONNRESET) {
      
      /* EHOSTUNREACH is the `No route to host' error that happens
	 probably when a PPP user shuts off his computer without
	 disconnecting cleanly.

	 ECONNRESET is the `Connection reset by peer' error that seems
	 to be a `normal' connection termination message.

	 Both messages are part of the normal operation of the system
	 and should only be logged as debug messages.  They happen
	 quite often.  If they are logged as errors they put lots of
	 meaningless messages in the syslog log file on production
	 machines. */

      Log_Hidden_System_Error (LOG_LEVEL_DEBUG0, __FILE__, __LINE__, errno,
			       "socket", "read",
			       "Connection was terminated incorrectly by "
			       "client");
      return FAILURE;
      
    } else {

      Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
				 "socket", "read", "CRD",
				 "Cannot read from client");
      return FAILURE;
    }
  } else if (num == READ_EOF) {
    Write_Message_To_Log (LOG_LEVEL_DEBUG1, __FILE__, __LINE__,
			  "EOF read from socket");
    return FAILURE;
  }

  /*
   * Calculate statistics
   */
  if (produce_flow_stats) {
    old_number_of_buffers_read = number_of_buffers_read;
    number_of_buffers_read++;
    average_read_buffer_size = (average_read_buffer_size
				* old_number_of_buffers_read
				+ num)
                                / number_of_buffers_read;
  }

  *number_of_bytes_read = num;
  return SUCCESS;
}

int 
Write_All_To_Socket (int file_descriptor, char *buffer_ptr,
		     int number_of_bytes_to_write)
/*
 * Functional Description:
 * 
 * This function writes all the bytes specified and does not return until it has
 * written all of them or an error occurs.
 * 
 * Formal Parameters:
 * 
 * file_descriptor (Read only) - descriptor to write to buffer_ptr (Read only) -
 * pointer to the buffer from which the bytes written will be taken
 * number_of_bytes_to_write (Read only) - number of bytes to write
 * 
 * Implicit Parameters:  socket_is_alive
 * 
 * Return Value:
 * 
 * None
 * 
 * Side Effects:  None
 * 
 */
{
  int             number_of_bytes_written = 0;
  int             total_bytes_written = 0;
  fd_set          output_set;
  int             select_status;
  struct timeval  timeout;

  timeout.tv_sec = 2;
  timeout.tv_usec = 0;

  while (total_bytes_written < number_of_bytes_to_write) {
    FD_ZERO (&output_set);
    FD_SET (file_descriptor, &output_set);

    select_status = select (file_descriptor + 1, (fd_set *) NULL,
			    &output_set, (fd_set *) NULL, &timeout);

    if (select_status == SELECT_ERROR) {
      Log_Hidden_System_Error (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       "unknown", "select",
			       "Could not wait for output readiness");
      return FAILURE;
    } else if (select_status == SELECT_TIMEOUT) {
      Write_Message_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, 
			    "Select timed out waiting for socket to be ready"
			    " to write");
      return FAILURE;
    }

    if (FD_ISSET (file_descriptor, &output_set)) {
      if (Write_To_Socket (file_descriptor, buffer_ptr,
			   number_of_bytes_to_write - total_bytes_written,
			   &number_of_bytes_written) == FAILURE)
	return FAILURE;

      buffer_ptr += number_of_bytes_written;
      total_bytes_written += number_of_bytes_written;
    } else {
      return FAILURE;
    }
  }  /* while */
  return SUCCESS;
}

int
Write_All_To_Socket_EonF (int fd, char *buffer_ptr, int number_to_write)
{
  if (Write_All_To_Socket (fd, buffer_ptr, number_to_write) == SUCCESS)
    return SUCCESS;
  exit (EXIT_FAILURE);
}

void 
Enable_Socket_Inactivity_Timeout (int sockdes)
{
  int             turn_it_on = 1;
  int             status;

  status = setsockopt (sockdes, SOL_SOCKET, SO_KEEPALIVE,
		       (char *) &turn_it_on, sizeof (turn_it_on));
  if (status == SETSOCKOPT_ERROR) {
    Write_System_Error_To_Log (LOG_LEVEL_CRIT, __FILE__, __LINE__, errno,
			       "socket", "setsockopt", "KPA",
			       "Could not enable socket inactivity"
			       " timeout");
    exit (EXIT_FAILURE);
  }
}



void 
Log_Socket_Stats ()
{
  Write_Va_Message_To_Log (LOG_LEVEL_INFO, __FILE__, __LINE__,
			   "Socket Stats: "
			   "Average Read Buffer Size = %f, "
			   "Number of EAGAIN Read Errors = %f, "
			   "Average Write Buffer Size = %f, "
			   "Number of EAGAIN Write Errors = %f",
			   average_read_buffer_size,
			   number_of_read_eagain_errors,
			   average_write_buffer_size,
			   number_of_write_eagain_errors);
}

void 
Start_Producing_Socket_Flow_Stats ()
{
  produce_flow_stats = 1;
}


/*
 * Purge_Socket_Input reads and discards any input currently in the kernel's
 * buffer for the file descriptor fd.  It performs a select to find out if
 * there is any input.  If there is no input a read is not done and the
 * function returns.  The function appears in this module so that
 * Read_from_Socket can be used to read the data.
 *
 * If an error is encountered, the function causes the whole program to exit.
 * This is because this function is used only in areas that this is
 * appropriate.  This is probably not a good general policy.
 */

void
Purge_Socket_Input (int fd)
{
  fd_set          input_set;
  struct timeval  timeout;
  int             n;

  timeout.tv_sec = 0;
  timeout.tv_usec = 0;
  FD_ZERO (&input_set);
  FD_SET (fd, &input_set);

  /*
   * Poll the descriptor once to find out if there is any input to get rid
   * of.
   */
  n = select (fd + 1, &input_set, (fd_set *) NULL, (fd_set *) NULL, &timeout);

  if (n == SELECT_ERROR) {
    Write_System_Error_To_Log (LOG_LEVEL_CRIT, __FILE__, __LINE__, errno,
			       "fd", "select", "INP",
			       "Error checking for input");
    exit (EXIT_FAILURE);
  }

  if (n == SELECT_TIMEOUT)
    return;

  if (Read_From_Socket (fd, module_buffer, MODULE_BUFFER_SIZE, &n) == FAILURE)
    exit (EXIT_FAILURE);

  Write_Va_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			   "read %d bytes of extraneous data from socket"
			   " input buffer", n);
}



int
Convert_h_errno_To_Text (int local_h_errno, char *h_errno_text)
{
  switch (local_h_errno) {
  case HOST_NOT_FOUND:
    STRNCPY(h_errno_text, "HOST_NOT_FOUND (man gethostbyaddr)");
    break;
  case TRY_AGAIN:
    STRNCPY(h_errno_text, "TRY_AGAIN (man gethostbyaddr)");
    break;
  case NO_RECOVERY:
    STRNCPY(h_errno_text, "NO_RECOVERY (man gethostbyaddr)");
    break;
  case NO_DATA:
    STRNCPY(h_errno_text, "NO_DATA (man gethostbyaddr)");
    break;
  default:
    STRNCPY(h_errno_text, "unknown error");
  }
}
